from __future__ import annotations
from datetime import date
from typing import Any, Dict, Optional, Union
from .generator import generate_id
from .governorates import get_governorate_name, get_region
from .utils import sanitize_id
from .validation import is_valid_id_str


class EgyptianNationalId:
    def __init__(self, id_value: Union[str, int]) -> None:
        """Create a new instance from a raw ID value.

        The value is converted to string and sanitized immediately.
        """
        self.id_str: str = sanitize_id(f"{id_value}")

    def __repr__(self) -> str:
        return f"<EgyptianNationalId {self.id_str}>"

    @staticmethod
    def sanitize(id_value: str) -> str:
        return sanitize_id(id_value)

    @classmethod
    def parse(cls, id_value: Union[str, int]) -> "EgyptianNationalId":
        return cls(id_value)

    @classmethod
    def is_valid_id(cls, id_value: Union[str, int]) -> bool:
        try:
            return is_valid_id_str(sanitize_id(f"{id_value}"))
        except Exception:
            return False

    @classmethod
    def check_is_male(cls, id_value: Union[str, int]) -> bool:
        inst = cls(id_value)
        return inst.is_valid() and inst.is_male()

    @classmethod
    def check_is_female(cls, id_value: Union[str, int]) -> bool:
        inst = cls(id_value)
        return inst.is_valid() and inst.is_female()

    @classmethod
    def check_is_adult(cls, id_value: Union[str, int]) -> bool:
        inst = cls(id_value)
        return inst.is_valid() and inst.is_adult()

    # --- Generator ---
    @classmethod
    def generate(cls, options: Optional[Dict[str, Any]] = None) -> str:
        return generate_id(options)

    def is_valid(self) -> bool:
        """Return True if the ID conforms to all validation rules."""
        return is_valid_id_str(self.id_str)

    def get_birth_year(self) -> int:
        century_digit = int(self.id_str[0])
        year = int(self.id_str[1:3])
        century = 1900 if century_digit == 2 else 2000
        return century + year

    def get_birth_month(self) -> int:
        return int(self.id_str[3:5])

    def get_birth_day(self) -> int:
        return int(self.id_str[5:7])

    def get_birth_date(self) -> Optional[date]:
        try:
            return date(
                self.get_birth_year(),
                self.get_birth_month(),
                self.get_birth_day(),
            )
        except ValueError:
            return None

    def get_age(self) -> Optional[int]:
        bd = self.get_birth_date()
        if bd is None:
            return None
        today = date.today()
        age = today.year - bd.year - ((today.month, today.day) < (bd.month, bd.day))
        return age

    # convenience property aliases
    @property
    def birth_date(self) -> Optional[date]:
        return self.get_birth_date()

    @property
    def age(self) -> Optional[int]:
        return self.get_age()

    @property
    def gender(self) -> str:
        return self.get_gender()

    @property
    def governorate(self) -> Optional[str]:
        return self.get_governorate_name()

    @property
    def region_name(self) -> str:
        return self.get_region()

    def get_governorate_code(self) -> str:
        return self.id_str[7:9]

    def get_governorate_name(self) -> Optional[str]:
        return get_governorate_name(self.get_governorate_code())

    def get_region(self) -> str:
        return get_region(self.get_governorate_code())

    def get_gender(self) -> str:
        return "female" if int(self.id_str[12]) % 2 == 0 else "male"

    def is_male(self) -> bool:
        return self.get_gender() == "male"

    def is_female(self) -> bool:
        return self.get_gender() == "female"

    def is_adult(self) -> bool:
        age = self.get_age()
        return age is not None and age >= 18

    def is_inside_egypt(self) -> bool:
        return self.get_governorate_code() != "88"

    def to_dict(self) -> Optional[Dict[str, Any]]:
        if not self.is_valid():
            return None
        return {
            "nationalId": self.id_str,
            "birthYear": self.get_birth_year(),
            "birthMonth": self.get_birth_month(),
            "birthDay": self.get_birth_day(),
            "age": self.get_age(),
            "gender": self.get_gender(),
            "governorate": self.get_governorate_name(),
            "region": self.get_region(),
            "insideEgypt": self.is_inside_egypt(),
            "isAdult": self.is_adult(),
        }



import unittest
from datetime import date
from egyid import EgyptianNationalId


class TestEgyptianNationalId(unittest.TestCase):
    def test_valid_id(self):
        # The last digit is a check digit, calculated using the standard Egyptian national ID algorithm.
        self.assertTrue(EgyptianNationalId.is_valid_id("29001011400014"))

    def test_invalid_id(self):
        self.assertFalse(EgyptianNationalId.is_valid_id("00000000000000"))

    def test_generate_valid(self):
        generated = EgyptianNationalId.generate({"year": 1990, "gender": "male", "governorate": "01"})
        self.assertTrue(EgyptianNationalId.is_valid_id(generated))


class TestEgyptianNationalIdComprehensive(unittest.TestCase):
    """Comprehensive test suite for Egyptian National ID validation and parsing."""

    def test_valid_ids(self):
        """Test various valid Egyptian National IDs."""
        valid_ids = [
            "29001011400014",  # 1990-01-01, Kalyubia, male
            "29010280165195",  # 1990-10-28, Cairo, male
            "30105201501238",  # 2001-05-20, Cairo, female
            "28801010100001",  # 1988-01-01, Cairo, male (adjusted checksum)
            "29512310123455",  # 1995-12-31, Cairo, female
        ]
        for id_str in valid_ids:
            with self.subTest(id=id_str):
                self.assertTrue(EgyptianNationalId.is_valid_id(id_str))
                eid = EgyptianNationalId(id_str)
                self.assertTrue(eid.is_valid())
                self.assertIsNotNone(eid.birth_date)
                self.assertIn(eid.gender, ['male', 'female'])
                self.assertIsNotNone(eid.governorate)

    def test_invalid_lengths(self):
        """Test IDs with wrong lengths."""
        invalid_lengths = [
            "123",           # Too short
            "2900101140001", # 13 digits
            "290010114000145", # 15 digits
            "",              # Empty
        ]
        for id_str in invalid_lengths:
            with self.subTest(id=id_str):
                self.assertFalse(EgyptianNationalId.is_valid_id(id_str))

    def test_invalid_century_digits(self):
        """Test IDs with invalid century digits."""
        invalid_centuries = [
            "19001011400014",  # Starts with 1
            "49001011400014",  # Starts with 4
            "09001011400014",  # Starts with 0
        ]
        for id_str in invalid_centuries:
            with self.subTest(id=id_str):
                self.assertFalse(EgyptianNationalId.is_valid_id(id_str))

    def test_invalid_dates(self):
        """Test IDs with invalid birth dates."""
        invalid_dates = [
            "29023011400014",  # Feb 30 (invalid)
            "29023111400014",  # Feb 31 (invalid)
            "29013211400014",  # Jan 32 (invalid)
            "29001311400014",  # Month 13 (invalid)
            "29010011400014",  # Month 00 (invalid)
            "29000111400014",  # Day 00 (invalid)
        ]
        for id_str in invalid_dates:
            with self.subTest(id=id_str):
                self.assertFalse(EgyptianNationalId.is_valid_id(id_str))

    def test_future_dates(self):
        """Test IDs with future birth dates."""
        # Assuming current year is 2026, future dates should be invalid
        future_ids = [
            "33601011400018",  # 2033-01-01 (future)
        ]
        for id_str in future_ids:
            with self.subTest(id=id_str):
                self.assertFalse(EgyptianNationalId.is_valid_id(id_str))

    def test_invalid_governorates(self):
        """Test IDs with invalid governorate codes."""
        invalid_govs = [
            "29001000400014",  # Gov 00 (invalid)
            "29001991400014",  # Gov 99 (invalid)
            "29001041400014",  # Gov 41 (invalid)
        ]
        for id_str in invalid_govs:
            with self.subTest(id=id_str):
                self.assertFalse(EgyptianNationalId.is_valid_id(id_str))

    def test_invalid_checksums(self):
        """Test IDs with wrong check digits."""
        invalid_checksums = [
            "29001011400015",  # Wrong checksum
            "29001011400013",  # Wrong checksum
            "29001011400010",  # Wrong checksum
        ]
        for id_str in invalid_checksums:
            with self.subTest(id=id_str):
                self.assertFalse(EgyptianNationalId.is_valid_id(id_str))

    def test_non_numeric_ids(self):
        """Test IDs with non-numeric characters."""
        non_numeric = [
            "2900101140001a",  # Letter
            "2900101140001 ",  # Space
            "2900101140001-",  # Dash
        ]
        for id_str in non_numeric:
            with self.subTest(id=id_str):
                self.assertFalse(EgyptianNationalId.is_valid_id(id_str))

    def test_arabic_numerals(self):
        """Test IDs with Arabic numerals (should be sanitized)."""
        arabic_id = "٢٩٠١٠٢٨٠١٦٥١٩٥"  # 29010280165195 in Arabic
        self.assertTrue(EgyptianNationalId.is_valid_id(arabic_id))

    def test_parsing_methods(self):
        """Test parsing methods for valid ID."""
        eid = EgyptianNationalId("29010280165195")
        self.assertEqual(eid.get_birth_year(), 1990)
        self.assertEqual(eid.get_birth_month(), 10)
        self.assertEqual(eid.get_birth_day(), 28)
        self.assertEqual(eid.birth_date, date(1990, 10, 28))
        self.assertEqual(eid.gender, "male")
        self.assertEqual(eid.governorate, "Cairo")
        self.assertEqual(eid.region_name, "Cairo")
        self.assertTrue(eid.is_male())
        self.assertFalse(eid.is_female())
        self.assertTrue(eid.is_adult())
        self.assertTrue(eid.is_inside_egypt())

    def test_generation(self):
        """Test ID generation."""
        generated = EgyptianNationalId.generate()
        self.assertTrue(EgyptianNationalId.is_valid_id(generated))
        self.assertEqual(len(generated), 14)

        # Test with options
        male_id = EgyptianNationalId.generate({"gender": "male", "year": 1990, "governorate": "01"})
        self.assertTrue(EgyptianNationalId.is_valid_id(male_id))
        eid = EgyptianNationalId(male_id)
        self.assertEqual(eid.get_birth_year(), 1990)
        self.assertEqual(eid.gender, "male")

    def test_to_dict(self):
        """Test dictionary output."""
        eid = EgyptianNationalId("29010280165195")
        data = eid.to_dict()
        expected = {
            "nationalId": "29010280165195",
            "birthYear": 1990,
            "birthMonth": 10,
            "birthDay": 28,
            "age": 35,  # As of 2025
            "gender": "male",
            "governorate": "Cairo",
            "region": "Cairo",
            "insideEgypt": True,
            "isAdult": True,
        }
        self.assertEqual(data, expected)

    def test_invalid_to_dict(self):
        """Test to_dict returns None for invalid IDs."""
        eid = EgyptianNationalId("00000000000000")
        self.assertIsNone(eid.to_dict())

    def test_class_methods(self):
        """Test class methods for quick checks."""
        self.assertTrue(EgyptianNationalId.check_is_male("29010280165195"))
        self.assertFalse(EgyptianNationalId.check_is_female("29010280165195"))
        self.assertTrue(EgyptianNationalId.check_is_adult("29010280165195"))

        self.assertFalse(EgyptianNationalId.check_is_male("30105201501220"))  # Female
        self.assertTrue(EgyptianNationalId.check_is_female("30105201501220"))

    def test_sanitization(self):
        """Test ID sanitization."""
        self.assertEqual(EgyptianNationalId.sanitize("29001011400014"), "29001011400014")
        self.assertEqual(EgyptianNationalId.sanitize("٢٩٠٠١٠١١٤٠٠٠١٤"), "29001011400014")
        self.assertEqual(EgyptianNationalId.sanitize("290-010-114-000-14"), "29001011400014")

    def test_edge_cases(self):
        """Test various edge cases."""
        # Leap year
        leap_id = "29602291400018"  # 1996-02-29
        self.assertTrue(EgyptianNationalId.is_valid_id(leap_id))

        # Non-leap year Feb 29
        invalid_leap = "29702291400014"  # 1997-02-29 (invalid)
        self.assertFalse(EgyptianNationalId.is_valid_id(invalid_leap))

        # Century boundary
        century_1900 = "29912311400019"  # 1999-12-31
        self.assertTrue(EgyptianNationalId.is_valid_id(century_1900))

        century_2000 = "30101011400014"  # 2001-01-01
        self.assertTrue(EgyptianNationalId.is_valid_id(century_2000))

    def test_foreign_governorate(self):
        """Test IDs from outside Egypt."""
        foreign_id = "29001018800018"  # Gov 88 (Outside Republic)
        eid = EgyptianNationalId(foreign_id)
        self.assertEqual(eid.governorate, "Outside the Republic")
        self.assertEqual(eid.region_name, "Foreign")
        self.assertFalse(eid.is_inside_egypt())


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import calendar
import random
from datetime import datetime
from typing import Any, Dict, Optional

from .governorates import GOVERNORATES
from .validation import compute_check_digit


def generate_id(options: Optional[Dict[str, Any]] = None) -> str:
    opts = options or {}

    year = opts.get("year", random.randint(1950, datetime.now().year))
    month = opts.get("month", random.randint(1, 12))
    max_day = calendar.monthrange(year, month)[1]
    day = opts.get("day", random.randint(1, max_day))

    if "governorate" in opts:
        gov = str(opts["governorate"]).zfill(2)
    else:
        gov = random.choice(list(GOVERNORATES.keys()))

    century_digit = 3 if year >= 2000 else 2
    year_digits = str(year)[-2:]
    month_digits = str(month).zfill(2)
    day_digits = str(day).zfill(2)

    if "gender" in opts:
        is_female = opts["gender"] == "female"
    else:
        is_female = bool(random.getrandbits(1))

    sequence_start = str(random.randint(0, 999)).zfill(3)
    gender_digit = random.randint(0, 4) * 2 if is_female else random.randint(0, 4) * 2 + 1

    id_without_check = (
        f"{century_digit}{year_digits}{month_digits}{day_digits}{gov}{sequence_start}{gender_digit}"
    )

    check_digit = compute_check_digit(id_without_check)
    return id_without_check + str(check_digit)


__all__ = ["generate_id"]



from __future__ import annotations

import re


def sanitize_id(id_value: str) -> str:
    """Return digits-only version of the ID, converting Arabic numerals."""
    trans = str.maketrans("٠١٢٣٤٥٦٧٨٩", "0123456789")
    cleaned = id_value.translate(trans)
    return re.sub(r"[^\d]", "", cleaned)


__all__ = ["sanitize_id"]


from __future__ import annotations

from typing import Dict, Optional, TypedDict


class GovernorateInfo(TypedDict):
    name: str
    region: str


GOVERNORATES: Dict[str, GovernorateInfo] = {
    "01": {"name": "Cairo", "region": "Cairo"},
    "02": {"name": "Alexandria", "region": "Alexandria"},
    "03": {"name": "Port Said", "region": "Canal"},
    "04": {"name": "Suez", "region": "Canal"},
    "11": {"name": "Damietta", "region": "Delta"},
    "12": {"name": "Dakahlia", "region": "Delta"},
    "13": {"name": "Ash Sharqia", "region": "Delta"},
    "14": {"name": "Kalyubia", "region": "Delta"},
    "15": {"name": "Kafr El Sheikh", "region": "Delta"},
    "16": {"name": "Gharbia", "region": "Delta"},
    "17": {"name": "Monufia", "region": "Delta"},
    "18": {"name": "Beheira", "region": "Delta"},
    "19": {"name": "Ismailia", "region": "Canal"},
    "21": {"name": "Giza", "region": "Upper Egypt"},
    "22": {"name": "Beni Suef", "region": "Upper Egypt"},
    "23": {"name": "Fayoum", "region": "Upper Egypt"},
    "24": {"name": "Minya", "region": "Upper Egypt"},
    "25": {"name": "Assiut", "region": "Upper Egypt"},
    "26": {"name": "Sohag", "region": "Upper Egypt"},
    "27": {"name": "Qena", "region": "Upper Egypt"},
    "28": {"name": "Aswan", "region": "Upper Egypt"},
    "29": {"name": "Luxor", "region": "Upper Egypt"},
    "31": {"name": "Red Sea", "region": "Frontier"},
    "32": {"name": "New Valley", "region": "Frontier"},
    "33": {"name": "Matrouh", "region": "Frontier"},
    "34": {"name": "North Sinai", "region": "Frontier"},
    "35": {"name": "South Sinai", "region": "Frontier"},
    "88": {"name": "Outside the Republic", "region": "Foreign"},
}


def get_governorate_info(code: str) -> Optional[GovernorateInfo]:
    return GOVERNORATES.get(code)


def get_governorate_name(code: str) -> Optional[str]:
    info = get_governorate_info(code)
    return info["name"] if info else None


def get_region(code: str) -> str:
    info = get_governorate_info(code)
    return info.get("region", "Foreign") if info else "Foreign"


__all__ = ["GovernorateInfo", "GOVERNORATES", "get_governorate_info", "get_governorate_name", "get_region"]


from __future__ import annotations
import re
from datetime import datetime
from .governorates import get_governorate_info


_MULTIPLIERS = (2, 7, 6, 5, 4, 3, 2, 7, 6, 5, 4, 3, 2)


def _is_valid_date(year: int, month: int, day: int) -> bool:
    try:
        datetime(year, month, day)
        return True
    except ValueError:
        return False


def _compute_check_digit(id_without_check: str) -> int:
    total = sum(int(id_without_check[i]) * _MULTIPLIERS[i] for i in range(13))
    remainder = total % 11
    if remainder == 0:
        return 0
    else:
        return (11 - remainder) % 10


def is_valid_id_str(id_str: str) -> bool:
    if not re.fullmatch(r"\d{14}", id_str):
        return False

    if int(id_str[0]) not in (2, 3):
        return False

    year = int(id_str[1:3])
    month = int(id_str[3:5])
    day = int(id_str[5:7])
    century = 1900 if int(id_str[0]) == 2 else 2000

    if not _is_valid_date(century + year, month, day):
        return False

    if century + year > datetime.now().year:
        return False

    if get_governorate_info(id_str[7:9]) is None:
        return False

    return (_compute_check_digit(id_str[:13]) == int(id_str[13]))


def compute_check_digit(id_without_check: str) -> int:
    return _compute_check_digit(id_without_check)


__all__ = ["is_valid_id_str", "compute_check_digit"]
